import math

def plan_plot_grid(n_plots: int) -> tuple:
    """find optimal placement (rows, cols) for given number of plots to place
    => min{rows - cols}  s.t. cols^2 <= n_plots <= rows^2 and cols * rows >= n_plots

    :param n_plots: number of plots to place
    :return: optimal (rows, cols)
    """
    rows = math.ceil(n_plots ** 0.5)
    cols = math.ceil(n_plots / rows)
    return rows, cols


def plot_walker(n_plots: int):
    """a generator which yields each step of row anc col number to place the plot,
    given total number of plots in the grid

    :param n_plots:
    :return: a generator of next plot loc (row and column)
    """
    rows, cols = plan_plot_grid(n_plots)  # get the placement grid size
    r, c = 0, 0
    for i in range(n_plots):
        yield r, c
        if c + 1 >= cols:
            # if current row is full, move to next row
            r += 1
            c = 0
        else:
            c += 1

def ax_walker(n_plots: int, axes):
    """a generator which yields each step of ax to place the plot

    :param n_plots: total number of plots in the grid
    :param axes: created by: fig, axes = plt.subplots(rows, cols, figsize = figsize)
    :return: a generator of next plot ax
    """
    rows, cols = plan_plot_grid(n_plots)
    if n_plots == 1:
        yield axes  # only single ax is contained in axes, return the single ax back

    elif rows == 1 or cols == 1:
        # it will still be a 1D array of ax, so only iterate along one dimension
        for ax in axes:
            yield ax
    else:
        # else its a 2D plot (have row and col > 1, the axes will be a 2D array)
        r, c = 0, 0
        for i in range(n_plots):
            yield axes[r, c]
            if c + 1 >= cols:
                # if current row is full, move to next row
                r += 1
                c = 0
            else:
                c += 1


from matplotlib import pyplot as plt


def plot_multiple_ys_against_x(
        ax,
        x,
        ys_dict: dict,
        ys_dict_r: dict = None,
        x_label: str = 'xs',
        y_label: str = 'y',
        y_label_r: str = 'y_r',
        title: str = "Untitled - Multiple Xs vs. Y",
        title_fontsize: str = "large",
        text_fontsize: str = "medium",
        show_grid: bool = True,
        show_legend: bool = True,
    ):
    """plot multiple Y values against same X-axis on the given ax

    :param ax: given ax to plot on, figure size is predefined in ax, fig
    :param x: array-like, x-axis values
    :param ys_dict: dictionary of several ys and their names {y_name : y_values} on the main (left) y_axis
    :param ys_dict_r: dictionary of several ys and their names {y_name : y_values} on the twin (right) y_axis
    :param x_label: label to show for x axis
    :param y_label: label to show for y axis on the main (left) ax
    :param y_label_r: label to show for y axis on the twin (right) ax
    :param title: title for the chart
    :param title_fontsize:
    :param text_fontsize:
    :param show_grid: whether to show grid
    :param show_legend: whether to show legend of different y names

    :return: ax of the plotted chart
    """

    for y_name, y in ys_dict.items():
        ax.plot(x, y, marker = 'o', label = y_name, lw = 2)

    ax.set_title(title, fontsize = title_fontsize)

    ax.set_xlabel(x_label, fontsize = text_fontsize)
    ax.set_ylabel(y_label, fontsize = text_fontsize)
    ax.tick_params(labelsize = text_fontsize)

    if show_grid:
        ax.grid(True)
    if show_legend:
        ax.legend()

    if ys_dict_r:
        #ax2_color = '#939b95'
        ax2 = ax.twinx()

        for y_name, y in ys_dict_r.items():
            ax2.plot(x, y, linestyle = ':', label = y_name, lw = 2)
       # ax2.plot(x, times, linestyle = ':', alpha = 0.8)
        ax2.set_ylabel(y_label_r, fontsize = text_fontsize)
        ax2.tick_params(labelsize = text_fontsize)


import logging
import jaydebeapi

from AqDsKit.accessors.db.drivers import download_jdbc_jar


class dfConfJdbc:
    def __init__(self, driver: str):
        self.driver = driver
        self.jar_path = None


    def bindServer(self, ip: str, port: int, db: str):
        self.ip = ip
        self.port = port
        self.db = db

    def login(self, username: str, password: str):
        self.username = username
        self.password = password

    def argTempStr(self):
        raise NotImplementedError("Must implement this method to get different arg placeholder for different database")

    def getConnStr(self) -> str:
        raise NotImplementedError("Must implement this method to get connection string")

    def getConnection(self) -> jaydebeapi.Connection:
        conn = jaydebeapi.connect(
            jclassname = self.driver,
            url = self.getConnStr(),
            driver_args = [self.username, self.password],
            jars = self.jar_path
        )
        return conn



class HiveJdbc(dfConfJdbc):
    def __init__(self):
        super().__init__("org.apache.hive.jdbc.HiveDriver")

    def argTempStr(self):
        return "?"

    def getConnStr(self) -> str:
        return f"jdbc:hive2://{self.ip}:{self.port}/{self.db}"

    def prepare(self, jar_folder):
        jar_file_url = "https://repo1.maven.org/maven2/org/apache/hive/hive-jdbc/3.1.2/hive-jdbc-3.1.2.jar"
        self.jar_path = download_jdbc_jar(jar_file_url, jar_folder)


class PrestoJdbc(dfConfJdbc):
    def __init__(self):
        super().__init__("com.facebook.presto.jdbc.PrestoDriver")

    def argTempStr(self):
        return "?"

    def getConnStr(self) -> str:
        return f"jdbc:presto://{self.ip}:{self.port}/{self.db}"

    def prepare(self, jar_folder):
        jar_file_url = "https://repo1.maven.org/maven2/io/prestosql/presto-jdbc/347/presto-jdbc-347.jar"
        self.jar_path = download_jdbc_jar(jar_file_url, jar_folder)


import logging
import os
from pyspark.sql import SparkSession
from pyspark import SparkContext, SparkConf


# os.environ['PYSPARK_PYTHON'] = 'python3.8'
# os.environ['PYSPARK_DRIVER_PYTHON'] = 'python3.8'
from AqDsKit.accessors.db.dbconfigs import baseDbInf
from AqDsKit.accessors.db.drivers import download_jdbc_jars
from AqDsKit.accessors.fs.s3 import S3Configs
from AqDsKit.utils.custom_structures import EnhancedDict


class SparkConnector:

    @ classmethod
    def setup(cls, s3_config: S3Configs, jdbc_driver_folder : str = os.getcwd()):
        """set up for the environment to urn successfully

        :param jdbc_driver_folder: folder path to save the jdbc drivers
        :return:
        """

        cls.STORAGE_URL = s3_config.endpoint_url
        cls.ACCESS_KEY = s3_config.aws_access_key_id
        cls.SECRET_ACCESS_KEY = s3_config.aws_secret_access_key
        cls.SESSION_TOKEN = s3_config.aws_session_token

        # download jdbc
        download_jdbc_jars(jdbc_driver_folder)
        cls.JAR_PATHS = ":".join(os.path.join(jdbc_driver_folder, jar) for jar in os.listdir(jdbc_driver_folder))
        logging.info(f"JDBC Driver Jars: {cls.JAR_PATHS}")

        ## build the configs
        cls.conf = EnhancedDict()
        cls.conf["spark.hadoop.fs.s3a.endpoint"] = cls.STORAGE_URL
        cls.conf["spark.hadoop.fs.s3a.aws.credentials.provider"] = 'org.apache.hadoop.fs.s3a.TemporaryAWSCredentialsProvider'
        cls.conf["spark.hadoop.fs.s3a.access.key"] = cls.ACCESS_KEY
        cls.conf["spark.hadoop.fs.s3a.secret.key"] = cls.SECRET_ACCESS_KEY
        cls.conf["spark.hadoop.fs.s3a.session.token"] = cls.SESSION_TOKEN
        cls.conf["spark.hadoop.fs.s3a.fast.upload"] = True
        cls.conf["spark.hadoop.fs.s3a.path.style.access"] = True
        cls.conf["spark.hadoop.fs.s3a.impl"] = "org.apache.hadoop.fs.s3a.S3AFileSystem"
        cls.conf["spark.hadoop.fs.s3a.multipart.size"] = "128M"
        cls.conf["spark.hadoop.fs.s3a.fast.upload.active.blocks"] = 8
        cls.conf["spark.driver.extraClassPath"] = cls.JAR_PATHS

    @ classmethod
    def getSparkSession(cls, app_name : str = "SparkConnector", master : str = "local"):

        sparkSessionBuilder = SparkSession.builder.master(master).appName(app_name)
        for k, v in cls.conf.items():
            sparkSessionBuilder = sparkSessionBuilder.config(k, v)

        spark = sparkSessionBuilder.getOrCreate()

        return spark

    def __init__(self, app_name : str = "SparkConnector", master : str = "local"):
        if not hasattr(self, 'JAR_PATHS'):
            raise Exception("Must call SparkConnector.setup() before initialization")

        self.spark = self.getSparkSession(app_name, master)
        logging.info(f"Spark session object successfully created, call SparkConnector({app_name}).spark to get spark obj")

    def get_file_path(self, path: str) -> str:
        return path

    def get_common_reader(self, header=True, sep=",", schema=None):
        if schema:
            return self.spark.read.option('header', header) \
                .option("inferSchema", "false") \
                .option("sep", sep) \
                .schema(schema)
        else:
            return self.spark.read.option('header', header) \
                .option("inferSchema", "true") \
                .option("sep", sep)

    def read_csv(self, *path: str, header=True, sep=",", schema=None):
        # path = self.get_file_path(path)

        try:
            df = self.get_common_reader(header, sep, schema).csv(*path)
        except Exception as e:
            logging.error('File Load FAILED: ' + str(e))
        else:
            logging.info('File Load successfully.')
            return df

    def read_sas(self, path: str):
        # path = self.get_file_path(path)

        try:
            df = self.spark.read.format("com.github.saurfang.sas.spark").load(path)
        except Exception as e:
            logging.error('File Load FAILED: ' + str(e))
        else:
            logging.info('File Load successfully.')
            return df

    def read_parquet(self, *path: str, header=True, sep=",", schema=None):
        # path = self.get_file_path(path)

        try:
            df = self.get_common_reader(header, sep, schema).parquet(*path)
        except Exception as e:
            logging.error('File Load FAILED: ' + str(e))
        else:
            logging.info('File Load successfully.')
            return df

    def save_csv(self, df, path: str, header=True, sep=",", mode="overwrite", repartition: int = None):
        path = self.get_file_path(path)

        if repartition:
            df = df.repartition(repartition)

        try:
            df.write.option('header', header).option("sep", sep).mode(mode).csv(path)
        except Exception as e:
            logging.error('File Wrote FAILED: ' + str(e))
        else:
            logging.info('File Wrote successfully.')

    def save_parquet(self, df, path: str, header=True, sep=",", mode="overwrite", repartition: int = None):
        path = self.get_file_path(path)

        if repartition:
            df = df.repartition(repartition)

        try:
            df.write.option('header', header).option("sep", sep).mode(mode).parquet(path)
        except Exception as e:
            logging.error('File Wrote FAILED: ' + str(e))
        else:
            logging.info('File Wrote successfully.')

    def query_db(self, dbConf: baseDbInf, sql: str, result_name: str = "temp_df", **kws):
        """query from database and return spark dataframe

        :param dbConf: the dbConf object for connection string setup
        :param sql: the sql query to make
        :param result_name: spark jdbc requires each query result to have a table name
        :param kws: other parameters that could parse into .jdbc(properties)
        :return: spark DataFrame
        """

        sql = f"({sql.replace(';', '')}) as {result_name}"  # drop ; and wrap with temp table name
        properties = {
            "driver": dbConf.getDriverClass(),
            # "dbtable" : sql,
        }

        if hasattr(dbConf, 'username'):
            properties["user"] = dbConf.username
            properties["password"] = dbConf.password

        for k, v in kws.items():
            properties[k] = v

        try:
            df = self.spark.read.jdbc(
                url=dbConf.getJdbcUrl(),
                table=sql,
                properties=properties
            )
        except Exception as e:
            logging.error('Query Load FAILED: ' + str(e))
        else:
            logging.info('Query Load successfully.')
            return df

    def save_db(self, dbConf: baseDbInf, df, table_name: str, mode: str = None, column_schema: dict = None,
                **kws) -> int:
        """save df to database

        :param dbConf: the dbConf object for connection string setup
        :param df: spark dataframe to save
        :param table_name: name of the table in the database
        :param mode: {None, append, overwrite}, None will fail if table exists, append will append to end, overwrite will discard existing values
        :param column_schema: dictionary of column and dtype, e.g. column_schema = {"cur" : "varchar(4)", "v1": "float"}
        :param kws: other parameters that could parse into .jdbc(properties)
        :return: 1 for success and 0 for failed
        """
        properties = {
            "driver": dbConf.getDriverClass(),
            "dbtable": table_name,
        }

        if hasattr(dbConf, 'username'):
            properties["user"] = dbConf.username
            properties["password"] = dbConf.password

        if column_schema:
            col_sch = ",".join(f"{k} {v}" for k, v in column_schema.items())
            properties["createTableColumnTypes"] = col_sch

        for k, v in kws.items():
            properties[k] = v

        try:
            df.write.jdbc(
                url=dbConf.getJdbcUrl(),
                table=table_name,
                mode=mode,
                properties=properties
            )
        except Exception as e:
            logging.error('Save to DB FAILED: ' + str(e))
            return 0
        else:
            logging.info('Save to DB successfully.')
            return 1


import matplotlib
import numpy as np
from cycler import cycler
from matplotlib import pyplot as plt
from sklearn.metrics import check_scoring

from AqDsKit.models.sklearn.hyperparam import train_different_hyperparam, cv_score_different_hyperparam
from AqDsKit.plots.commons.basics import plot_multiple_ys_against_x
from AqDsKit.plots.commons.tools import plan_plot_grid, ax_walker

""" plotting 
"""


def plot_cv_scores_over_hyperparams(
        clf,
        X,
        y = None,
        param_grid: dict = None,
        scoring_grid: dict = None,
        absolute_score: bool = False,
        cv: int = 5,
        n_jobs: int = -1,
        title: str = 'Score Over Hyperparameter',
        ax: matplotlib.axes.Axes = None,
        figsize: tuple = None,
        title_fontsize: str = "large",
        text_fontsize: str = "medium"
    ):
    """Plots score(s) of a model over for specified hyperparameter(s)

    Args:
        clf: Clusterer instance that implements ``fit``,``fit_predict``, and
            ``score`` methods (or have labels_ attribute), and an ``n_clusters`` or ``n_components`` hyperparameter.
            e.g. :class:`sklearn.cluster.KMeans` instance, GaussianMixture instance

        X (array-like, shape (n_samples, n_features)):
            Data to cluster, where n_samples is the number of samples and
            n_features is the number of features.

        y: y_train to give the label info, None for unsupervised learning

        param_grid: a dictionary of {param_name, param_ranges}, you can add multiple params to test, and will create one plot per param
            param_name: name of the hyperparameter
            param_ranges: (None or :obj:`list` of int, optional): List of
                param values for which to train the model and plot the corresponding score

        scoring_grid: dictionary of {score_name : scoring_func}, you can add multiple score metrics
            score_name: name of the score to be displayed on the plot
            scoring_func: if given, use this scoring func to calculate score, otherwise plese parse None and it will use model.score()
                            scoring_func can be either str (sklearn pre-defined scorer) or callable (custom scorer func)

        absolute_score: whether the score need to be a positive number (sometimes scores are negative, e.g. score of K-means)

        cv: number of cross validation circles; set to -1 or None if you want to skip CV and only do on training set

        n_jobs (int, optional): Number of jobs to run in parallel. Defaults to -1.

        title (string, optional): Title of the generated plot

        ax (:class:`matplotlib.axes.Axes`, optional): The axes upon which to
            plot the curve. If None, the plot is drawn on a new set of axes.

        figsize (2-tuple, optional): Tuple denoting figure size of the plot
            e.g. (6, 6). Defaults to ``None``.

        title_fontsize (string or int, optional): Matplotlib-style fontsizes.
            Use e.g. "small", "medium", "large" or integer-values. Defaults to
            "large".

        text_fontsize (string or int, optional): Matplotlib-style fontsizes.
            Use e.g. "small", "medium", "large" or integer-values. Defaults to
            "medium".

    Returns:
        ax (:class:`matplotlib.axes.Axes`): The axes on which the plot was
            drawn.
    """

    # if scoring_grid is not passed, use single default score method (set scoring_func to None):
    if scoring_grid is None:
        scoring_grid = {'score' : None}

    if param_grid is None:
        raise ValueError("Must specify at least one param to plot via param_grid attribute: param_grid = {param_name : param_ranges}")

    param_scorings = {} # save {param_name : {score_name: scores} } for each scoring metric and each param_name

    for param_name, param_ranges in param_grid.items():
        scorings = {} # save {score_name: scores} for each scoring metric

        if cv == -1 or cv is None:
            # skip cross validation and build score only on the training set (full X provided)
            clfs, times = train_different_hyperparam(clf, X, y, param_name = param_name, param_ranges = param_ranges,
                                                     n_jobs = n_jobs)

            for score_name, scoring_func in scoring_grid.items():
                # if scoring_func is None:
                #     scores = [clf.score(X, y) for clf in clfs]
                # else:

                scoring_func = check_scoring(clf, scoring_func) #  scoring_func can be str or callable (custom scorer func)
                scores = [scoring_func(clf, X, y) for clf in clfs]  # for K-means, score = -inertia

                if absolute_score:
                    scores = np.absolute(scores)

                scorings[score_name] = scores
        else:
            # otherwise use cross validation
            # will not train during the model build
            for score_name, scoring_func in scoring_grid.items():
                # parallel run cv_score process as it will take a lot of time
                scores, times = cv_score_different_hyperparam(clf, X, y, param_name = param_name, param_ranges = param_ranges,
                                                              scoring_func = scoring_func, cv = cv, n_jobs = n_jobs)
                #scores = [cross_val_score(clf, X, y, scoring = scoring_func, cv = cv, n_jobs = n_jobs).mean() for clf in clfs]
                if absolute_score:
                    scores = np.absolute(scores)

                scorings[score_name] = scores

        param_scorings[param_name] = scorings

    # decide the size/layout of the plot
    n_plots = len(param_grid)
    rows, cols = plan_plot_grid(n_plots)

    if ax is None:
        fig, axes = plt.subplots(rows, cols, figsize = figsize)
    else:
        axes = ax

    walker = ax_walker(n_plots, axes)
    plt.rc('axes', prop_cycle = (cycler('color', ['#7969aa', '#678659', '#9d9962', '#b35252', '#a26fa6'])))

    for ax, (param_name, scorings) in zip(walker, param_scorings.items()):

        sub_title = f"{title}\n({param_name})"
        param_ranges = param_grid[param_name]

        plot_multiple_ys_against_x(
            ax, x = param_ranges, ys_dict = scorings, ys_dict_r = None,
            x_label = param_name, y_label = 'Score',
            title = sub_title, title_fontsize = title_fontsize, text_fontsize = text_fontsize,
        )


    fig.tight_layout()

    return axes





import time

from sklearn.model_selection import cross_val_score

from AqDsKit.models.sklearn.general import clone_set_hyperparam
from AqDsKit.utils.parallel import delayer, parallel_run


@ delayer
def _set_hyperparam_train(clf, X, y = None, param_name: str = None, param_value = None):
    """Clone clf, set given hyperparameter and train the model.

    :param clf: Clusterer instance that implements ``fit`` method, and an ``n_clusters`` or ``n_components`` hyperparameter.
                    e.g. :class:`sklearn.cluster.KMeans` instance, GaussianMixture instance
    :param X: (array-like, shape (n_samples, n_features)):
                    Data to cluster, where n_samples is the number of samples and
                    n_features is the number of features.
    :param y: y to be fed into the training
    :param param_name: name of the hyperparameter, must be an attribute of clf
    :param param_value: value of the hyperparameter to be set
    :return: fitted clf and time of training
    """

    start = time.time()
    clf = clone_set_hyperparam(clf, param_name, param_value)
    return clf.fit(X, y), time.time() - start

def train_different_hyperparam(clf, X, y = None, param_name: str = None, param_ranges: list = None, n_jobs: int = -1):
    # use parallel run to boost speed, tuples = [(clf1, time1), (clf2, time2), ... (clfn, timen)] where n = len(param_ranges)

    # equivalent to:
    # tuples = Parallel(n_jobs = n_jobs)(delayed(_set_hyperparam_train)(clf, X, y, param_name, param) for param in param_ranges)
    jobs = (_set_hyperparam_train(clf, X, y, param_name, param) for param in param_ranges)
    tuples = parallel_run(jobs, n_jobs = n_jobs)

    clfs, times = zip(*tuples)
    return clfs, times



@ delayer
def _set_hyperparm_cv_score(clf, X, y = None, param_name: str = None, param_value = None, scoring_func = None, cv = 5, n_jobs = -1):
    start = time.time()
    clf = clone_set_hyperparam(clf, param_name, param_value)
    cv_score = cross_val_score(clf, X, y, scoring = scoring_func, cv = cv, n_jobs = n_jobs).mean(),
    return  cv_score, time.time() - start

def cv_score_different_hyperparam(clf, X, y, param_name: str = None, param_ranges: list = None, scoring_func = None, cv: int = 5, n_jobs: int = -1):
    # use parallel run to boost speed, tuples = [(score1, time1), (score2, time2), ... (scoren, timen)] where n = len(param_ranges)

    # equivalent to:
    # tuples = Parallel(n_jobs = n_jobs)(delayed(_set_hyperparm_cv_score)(clf, X, y, param_name, param, scoring_func, cv, n_jobs) for param in param_ranges)
    jobs = (_set_hyperparm_cv_score(clf, X, y, param_name, param, scoring_func, cv, n_jobs) for param in param_ranges)
    tuples = parallel_run(jobs, n_jobs = n_jobs)

    scores, times = zip(*tuples)
    return scores, times


#######
# To be compatible with sklearn
# https://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter
# scoring takes y_true and y_pred and returns a single score
#######

import numpy as np

# example of custom scoring metric
def max_diff(y_true, y_pred, **kws):
    diff = np.abs(y_true - y_pred).max()
    return np.log1p(diff)



import matplotlib
from matplotlib import pyplot as plt
import numpy as np
from scikitplot.metrics import plot_silhouette

from AqDsKit.plots.commons.tools import plan_plot_grid, ax_walker
from AqDsKit.plots.sklearn.general import plot_cv_scores_over_hyperparams, train_different_hyperparam

""" 
Visually Choose the best n_clusters for the clustering model (silhouette analysis and elbow method) 
"""

def plot_elbow_curve(
        clf,
        X,
        cluster_ranges: list = None,
        n_jobs: int = -1,
        title = 'Elbow Plot',
        ax: matplotlib.axes.Axes = None,
        figsize: tuple = None,
        title_fontsize: str = "large",
        text_fontsize: str = "medium"
    ):
    """Plots elbow curve of different values of K for KMeans clustering.

    Args:
        clf: Clusterer instance that implements ``fit``,``fit_predict``, and
            ``score`` methods (or have labels_ attribute), and an ``n_clusters`` or ``n_components`` hyperparameter.
            e.g. :class:`sklearn.cluster.KMeans` instance, GaussianMixture instance

        X (array-like, shape (n_samples, n_features)):
            Data to cluster, where n_samples is the number of samples and
            n_features is the number of features.

        cluster_ranges (None or :obj:`list` of int, optional): List of
            n_clusters for which to plot the explained variances. Defaults to
            ``range(1, 12, 2)``.

        n_jobs (int, optional): Number of jobs to run in parallel. Defaults to -1.

        title (string, optional): Title of the generated plot. Defaults to "Elbow Plot"

        ax (:class:`matplotlib.axes.Axes`, optional): The axes upon which to
            plot the curve. If None, the plot is drawn on a new set of axes.

        figsize (2-tuple, optional): Tuple denoting figure size of the plot
            e.g. (6, 6). Defaults to ``None``.

        title_fontsize (string or int, optional): Matplotlib-style fontsizes.
            Use e.g. "small", "medium", "large" or integer-values. Defaults to
            "large".

        text_fontsize (string or int, optional): Matplotlib-style fontsizes.
            Use e.g. "small", "medium", "large" or integer-values. Defaults to
            "medium".

    Returns:
        ax (:class:`matplotlib.axes.Axes`): The axes on which the plot was
            drawn.

    Example:
        >>> kmeans = KMeans(random_state=1)
        >>> plot_elbow_curve(kmeans, cluster_ranges=range(1, 30))
        <matplotlib.axes._subplots.AxesSubplot object at 0x7fe967d64490>
        >>> plt.show()
    """

    if cluster_ranges is None:
        cluster_ranges = range(1, 12, 2)
    else:
        cluster_ranges = sorted(cluster_ranges)


    scoring_grid = {'Sum of Squared Errors' : None}

    if hasattr(clf, 'n_clusters'):
        param_grid = {'n_clusters' : cluster_ranges}

    elif hasattr(clf, 'n_components'):
        param_grid = {'n_components' : cluster_ranges}

    else:
        raise TypeError('"n_clusters" or "n_components" attribute not in classifier. '
                        'Cannot plot elbow method.')

    plot_cv_scores_over_hyperparams(
        clf, X, y = None, param_grid = param_grid, scoring_grid = scoring_grid,
        absolute_score = True, n_jobs = n_jobs, title = title,
        ax = ax, figsize = figsize,
        title_fontsize = title_fontsize, text_fontsize = text_fontsize
    )


def plot_silhouettes(
        clf,
        X,
        cluster_ranges: list = None,
        n_jobs: int = -1,
        title = 'Silhouette Analysis',
        metric = 'euclidean',
        copy: bool = True,
        figsize: tuple = None,
        cmap = 'tab20b',
        title_fontsize = "large",
        text_fontsize = "medium"
    ):
    """
    Args:
        clf: Clusterer instance that implements ``fit``,``fit_predict``, and
            ``score`` methods (or have labels_ attribute), and an ``n_clusters`` or ``n_components`` hyperparameter.
            e.g. :class:`sklearn.cluster.KMeans` instance, GaussianMixture instance

        X (array-like, shape (n_samples, n_features)):
            Data to cluster, where n_samples is the number of samples and
            n_features is the number of features.

        cluster_ranges (array-like, shape (n_samples,)):
            Cluster label for each sample.

        n_jobs (int, optional): Number of jobs to run in parallel. Defaults to -1.

        title (string, optional): Title of the generated plot. Defaults to
            "Silhouette Analysis"

        metric (string or callable, optional): The metric to use when
            calculating distance between instances in a feature array.
            If metric is a string, it must be one of the options allowed by
            sklearn.metrics.pairwise.pairwise_distances. If X is
            the distance array itself, use "precomputed" as the metric.

        copy (boolean, optional): Determines whether ``fit`` is used on
            **clf** or on a copy of **clf**.

        figsize (2-tuple, optional): Tuple denoting figure size of the plot
            e.g. (6, 6). Defaults to ``None``.

        cmap (string or :class:`matplotlib.colors.Colormap` instance, optional):
            Colormap used for plotting the projection. View Matplotlib Colormap
            documentation for available options.
            https://matplotlib.org/users/colormaps.html

        title_fontsize (string or int, optional): Matplotlib-style fontsizes.
            Use e.g. "small", "medium", "large" or integer-values. Defaults to
            "large".

        text_fontsize (string or int, optional): Matplotlib-style fontsizes.
            Use e.g. "small", "medium", "large" or integer-values. Defaults to
            "medium".
    """
    if cluster_ranges is None:
        cluster_ranges = range(1, 12, 2)
    else:
        cluster_ranges = sorted(cluster_ranges)

    if hasattr(clf, 'n_clusters'):
        clfs, times = train_different_hyperparam(clf, X, y = None, param_name ='n_clusters',
                                                 param_ranges = cluster_ranges, n_jobs = n_jobs)

    elif hasattr(clf, 'n_components'):
        clfs, times = train_different_hyperparam(clf, X, y = None, param_name ='n_components',
                                                 param_ranges = cluster_ranges, n_jobs = n_jobs)

    else:
        raise TypeError('"n_clusters" or "n_components" attribute not in classifier. '
                        'Cannot plot multiple silhouettes')


    rows, cols = plan_plot_grid(len(cluster_ranges))

    fig, axes = plt.subplots(rows, cols, figsize=figsize)
    walker = ax_walker(len(cluster_ranges), axes)
    fig.suptitle(title)

    for ax, k, clf in zip(walker, cluster_ranges, clfs):
        if hasattr(clf, 'labels_'):
            cluster_labels = clf.labels_  # cluster must have labels_ after training
        else:
            cluster_labels = clf.predict(X)

        plot_silhouette(X, cluster_labels, title = f'$k={k}$',
                    metric = metric, copy = copy, ax = ax,
                    cmap = cmap, title_fontsize = title_fontsize,
                    text_fontsize = text_fontsize)


#######
# To be compatible with sklearn
# https://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter
# scorer is a callable object takes estimator, X, y_true and returns a score
# scorer can be used in Gridsearch to automatically find optimal hyperparamters
# scorer can also be used standalone, scorer(model, X_test, y_test) to get a score
#######
from sklearn.metrics import make_scorer

#################### scorer for supervised learning ######################
from AqDsKit.metrics.sklearn.custom_scoring import max_diff

max_diff_scorer = make_scorer(max_diff, greater_is_better = False, needs_proba = False, needs_threshold = False)


#################### scorer for clustering ######################

from sklearn.metrics import silhouette_score

def silhouette_scorer(estimator, X, y_true = None) -> float:
    """can be used to help select best number of clusters while tuning hyperparams for clustering problems
    suitable for K-means, DBSCAN(with predict method implemented), etc.

    :param estimator: must have ``predict`` method which returns predicting class
    :param X:
    :param y_true: Not used
    :return: silhouette score of a given estimation
    """
    if not callable(getattr(estimator, "predict", None)):
        raise TypeError(f"This estimator: {estimator} does not have ``predict`` method, does not support silhouette scoring")
    labels = estimator.predict(X)
    score = silhouette_score(X, labels)
    return score


def aic_scorer(estimator, X, y_true = None) -> float:
    """AIC score for Gaussian Mixture Model Metric
    can be used in GridSearch for hyperparameter tuning

    :param estimator: must have ``aic`` method which returns aic score
    :param X:
    :param y_true:
    :return: the protocol requires greater the better, so it will return negative aic score
    """
    if not callable(getattr(estimator, "aic", None)):
        raise TypeError(f"This estimator: {estimator} does not have ``aic`` method, does not support aic metric")

    aic = estimator.aic(X)
    return -aic


def bic_scorer(estimator, X, y_true = None) -> float:
    """BIC score for Gaussian Mixture Model Metric
    can be used in GridSearch for hyperparameter tuning

    :param estimator: must have ``bic`` method which returns bic score
    :param X:
    :param y_true:
    :return: the protocol requires greater the better, so it will return negative bic score
    """
    if not callable(getattr(estimator, "bic", None)):
        raise TypeError(f"This estimator: {estimator} does not have ``aic`` method, does not support aic metric")

    bic = estimator.bic(X)
    return -bic


import logging
import pandas as pd
import numpy as np
import contextlib


from sqlalchemy import create_engine, and_
from sqlalchemy.orm import sessionmaker, Session, Query
from sqlalchemy.inspection import inspect


#######################################
###  Base Database Configuration    ###
#######################################

class baseDbInf:
    def __init__(self, driver: str):
        self.driver = driver

    def bindServer(self, ip: str, port: int, db: str):
        self.ip = ip
        self.port = port
        self.db = db

    def login(self, username: str, password: str):
        self.username = username
        self.password = password

    def argTempStr(self):
        raise NotImplementedError("Must implement this method to get different arg placeholder for different database")

    def getConnStr(self) -> str:
        #engine = create_engine('mysql+mysqlconnector://USRNAME:PSWD@localhost:3306/DATABASE?charset=ytf8')
        #engine = create_engine("ibm_db_sa://USRNAME:PSWD@IP:PORT/DATABASE?charset=utf8")
        #engine = create_engine('sqlite:///DB_ADDRESS')
        return f"{self.driver}://{self.username}:{self.password}@{self.ip}:{self.port}/{self.db}?charset=utf8"

    def launch(self):
        connStr = self.getConnStr()
        self.engine = create_engine(connStr)
        self.DBSession = sessionmaker(bind = self.engine)
        logging.info("Engine started, ready to go!")

    def newSession(self):
        try:
            session = self.DBSession()
        except Exception as e:
            logging.error(e)
        else:
            return session

    def getJdbcUrl(self) -> str:
        # for spark connection and other purpose
        raise NotImplementedError("Must implement this method to get different JdbcUrl for different database")

    def getDriverClass(self) -> str:
        # for spark connection and other purpose
        raise NotImplementedError("Must implement this method to get different DriverClass for different database")


#############################################
### Provider level database Configuration ###
#############################################

class MySQL(baseDbInf):
    def __init__(self, driver = "mysqlconnector"):
        super().__init__(f"mysql+{driver}")

    def argTempStr(self):
        return "%s"

    def getJdbcUrl(self) -> str:
        return f"jdbc:mysql://{self.ip}:{self.port}/{self.db}"

    def getDriverClass(self) -> str:
        # https://repo1.maven.org/maven2/mysql/mysql-connector-java/8.0.27/mysql-connector-java-8.0.27.jar
        return "com.mysql.jdbc.Driver"

class DB2(baseDbInf):
    def __init__(self):
        super().__init__("ibm_db_sa")

    def argTempStr(self):
        return "?"

    def getJdbcUrl(self) -> str:
        return f"jdbc:db2://{self.ip}:{self.port}/{self.db}"

    def getDriverClass(self) -> str:
        # https://repo1.maven.org/maven2/com/ibm/db2/jcc/db2jcc/db2jcc4/db2jcc-db2jcc4.jar
        return "com.ibm.db2.jcc.DB2Driver"

class SqLite(baseDbInf):
    def __init__(self, dburl):
        super().__init__("sqlite")
        self.dburl = dburl

    def getConnStr(self) -> str:
        return f"sqlite:///{self.dburl}"

    def argTempStr(self):
        return "?"

    def getJdbcUrl(self) -> str:
        return f"jdbc:sqlite://{self.dburl}"

    def getDriverClass(self) -> str:
        # https://repo1.maven.org/maven2/org/xerial/sqlite-jdbc/3.36.0.3/sqlite-jdbc-3.36.0.3.jar
        return "org.sqlite.JDBC"

class SqlServer(baseDbInf):
    def __init__(self):
        super().__init__("mssql+pymssql")

    def argTempStr(self):
        return "?"

    def getJdbcUrl(self) -> str:
        return f"jdbc:sqlserver://{self.ip}:{self.port}/{self.db}"

    def getDriverClass(self) -> str:
        # https://repo1.maven.org/maven2/com/microsoft/sqlserver/mssql-jdbc/8.4.0.jre8/mssql-jdbc-8.4.0.jre8.jar
        return "com.microsoft.sqlserver.jdbc.SQLServerDriver"

class Oracle(baseDbInf):
    def __init__(self, sid: str):
        super().__init__("oracle")
        self.sid = sid

    def argTempStr(self):
        return "?"

    def getJdbcUrl(self) -> str:
        return f"jdbc:oracle:thin:/@(DESCRIPTION=(ADDRESS=(PROTOCOL=tcp)(HOST={self.ip})(PORT={self.port}))(CONNECT_DATA=(SID={self.sid})))"

    def getDriverClass(self) -> str:
        # https://repo1.maven.org/maven2/com/oracle/database/jdbc/ojdbc8/19.3.0.0/ojdbc8-19.3.0.0.jar
        return "oracle.jdbc.driver.OracleDriver"

class PostgresSql(baseDbInf):
    def __init__(self):
        super().__init__("postgresql+psycopg2")

    def argTempStr(self):
        return "?"

    def getJdbcUrl(self) -> str:
        return f"jdbc:postgresql://{self.ip}:{self.port}/{self.db}"

    def getDriverClass(self) -> str:
        # https://repo1.maven.org/maven2/org/postgresql/postgresql/42.2.18/postgresql-42.2.18.jar
        return "org.postgresql.Driver"


import json
import logging
import os
from collections import OrderedDict



''' ---- 			file loader 			---- '''

def loadJSON(file):
    if isinstance(file, (list, dict)):
        return file
    elif isinstance(file, str):
        with open(file, "rb") as obj:
            return json.load(obj, object_pairs_hook = OrderedDict)
    else:
        raise ValueError("Please parse a file path or JS object")


def toJSON(js, file):
    with open(file, "w") as obj:
        json.dump(js, obj, indent = 4)



class File:
    def __init__(self, fileModule):
        self.__fileModule = fileModule

    def getDirPath(self):
        return os.path.abspath(os.path.dirname(self.__fileModule))

    def concatPath(self, relativefileName):
        paths = relativefileName.split("/")
        return os.path.join(self.getDirPath(), *paths)

    def loadJson(self, fileName):
        filepath = self.concatPath(fileName)

        file = loadJSON(filepath)
        return file

    def toJson(self, df, fileName):
        filepath = self.concatPath(fileName)
        toJSON(df, filepath)

    def loadCsv(self, fileName):
        filepath = self.concatPath(fileName)
        file = pd.read_csv(filepath, encoding="utf8", engine="python")
        return file

    def loadYML(self, fileName):
        filepath = self.concatPath(fileName)
        with open(filepath, "rb") as obj:
            cfg = yaml.load(obj)
        return cfg


from sklearn import clone


def clone_set_hyperparam(clf, param_name: str = None, param_value = None):
    """Clone clf, set given hyperparameter and return the model.

    """
    clf = clone(clf)
    if hasattr(clf, param_name):
        setattr(clf, param_name, param_value)
    else:
        raise TypeError(f"estimator {clf} has no attribute {param_name}")

    return clf


import logging
import os

import requests


def download_jdbc_jar(jar_file_url, jdbc_driver_folder, jar_file_name: str = None):
    if not os.path.exists(jdbc_driver_folder):
        os.makedirs(jdbc_driver_folder)

    if not jar_file_name:
        jar_file_name = jar_file_url.split('/')[-1]

    jar_path = os.path.join(jdbc_driver_folder, jar_file_name)

    proxies = {
        'http': os.getenv('HTTP_PROXY', None),
        'https': os.getenv('HTTPS_PROXY', None)
    }

    if proxies['http']:
        with requests.get(
                jar_file_url,
                proxies = proxies
        ) as response, open(jar_path, 'wb') as out_file:
            out_file.write(response.content)
    else:
        with requests.get(jar_file_url) as response, open(jar_path, 'wb') as out_file:
            out_file.write(response.content)

    logging.info(f"Successfully download JDBC jar from {jar_file_url} and saved to {jar_path}")
    return jar_path


def download_jdbc_jars(jdbc_driver_folder):
    jars = [
        {
            'db_provider': 'db2',
            'jar_file_url': 'https://repo1.maven.org/maven2/com/ibm/db2/jcc/db2jcc/db2jcc4/db2jcc-db2jcc4.jar',
            'jar_file_name': 'db2jcc-db2jcc4.jar'
        },
        {
            'db_provider': 'mysql',
            'jar_file_url': 'https://repo1.maven.org/maven2/mysql/mysql-connector-java/8.0.27/mysql-connector-java-8.0.27.jar',
            'jar_file_name': 'mysql-connector-java.jar'
        },
        {
            'db_provider': 'sqlite',
            'jar_file_url': 'https://repo1.maven.org/maven2/org/xerial/sqlite-jdbc/3.36.0.3/sqlite-jdbc-3.36.0.3.jar',
            'jar_file_name': 'sqlite-jdbc.jar'
        },
        {
            'db_provider': 'sqlserver',
            'jar_file_url': 'https://repo1.maven.org/maven2/com/microsoft/sqlserver/mssql-jdbc/8.4.0.jre8/mssql-jdbc-8.4.0.jre8.jar',
            'jar_file_name': 'mssql-jdbc-8.4.0.jre8.jar'
        },
        {
            'db_provider': 'oracle',
            'jar_file_url': 'https://repo1.maven.org/maven2/com/oracle/database/jdbc/ojdbc8/19.3.0.0/ojdbc8-19.3.0.0.jar',
            'jar_file_name': 'ojdbc.jar'
        },
        {
            'db_provider': 'presto',
            'jar_file_url': 'https://repo1.maven.org/maven2/io/prestosql/presto-jdbc/347/presto-jdbc-347.jar',
            'jar_file_name': 'presto.jar'
        },
        {
            'db_provider': 'postgresql',
            'jar_file_url': 'https://repo1.maven.org/maven2/org/postgresql/postgresql/42.2.18/postgresql-42.2.18.jar',
            'jar_file_name': 'postgresql.jar'
        },
        {
            'db_provider': 'sas',
            'jar_file_url': 'https://repos.spark-packages.org/saurfang/spark-sas7bdat/3.0.0-s_2.12/spark-sas7bdat-3.0.0-s_2.12.jar',
            'jar_file_name': 'spark-sas.jar'
        },
        {
            'db_provider': 'parso',  # use to read sas dataset
            'jar_file_url': 'https://repo1.maven.org/maven2/com/epam/parso/2.0.14/parso-2.0.14.jar',
            'jar_file_name': 'parso.jar'
        }
    ]

    if not os.path.exists(jdbc_driver_folder):
        os.makedirs(jdbc_driver_folder)

    for jar in jars:
        db_provider = jar.get('db_provider')
        jar_file_url = jar.get('jar_file_url')
        jar_file_name = jar.get('jar_file_name')

        download_jdbc_jar(jar_file_url, jdbc_driver_folder, jar_file_name)


import numpy as np
from sklearn.base import ClassifierMixin, BaseEstimator
from sklearn.cluster import DBSCAN
from sklearn.neighbors import KNeighborsClassifier
from sklearn.utils.validation import check_is_fitted, _check_sample_weight, _deprecate_positional_args


class PredictableDBSCAN(DBSCAN):
    """DBSCAN in sklearn does not have predict() method, so this version implements predict method
    """

    @_deprecate_positional_args
    def __init__(
            self,
            eps = 0.5,
            *,
            min_samples = 5,
            metric = 'euclidean',
            metric_params = None,
            algorithm = 'auto',
            leaf_size = 30,
            p = None,
            n_jobs = None,
            clf: BaseEstimator = KNeighborsClassifier(),
        ):
        """
        parameters can be checked here:
        https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html#sklearn.cluster.DBSCAN

        :param clf: sklearn implementation of classifier instance, by default it use KNN
        :param eps:
        :param min_samples:
        :param metric:
        :param metric_params:
        :param algorithm:
        :param leaf_size:
        :param p:
        :param n_jobs:
        """
        super().__init__(eps = eps, min_samples=min_samples, metric=metric,
                 metric_params=metric_params, algorithm=algorithm, leaf_size=leaf_size, p=p,
                 n_jobs=n_jobs)
        self.clf = clf

    def fit(self, X, y = None, sample_weight = None):
        """Will train the custom classifier on the output of DBSCAN

        :param X:
        :param y: not used
        :param sample_weight:
        :return:
        """
        super().fit(X, y, sample_weight = sample_weight)
        self.clf.fit(self.components_, self.labels_[self.core_sample_indices_])
        return self


    def predict(self, X, sample_weight=None):
        """Predict the closest cluster each sample in X belongs to.

        A custom classifier is trained on the components_ (X) and predictions of the DBSCAN labels_ (y)
        And this trained classifier is used to predict the labels

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            New data to predict.

        sample_weight : array-like of shape (n_samples,), default=None
            The weights for each observation in X. If None, all observations
            are assigned equal weight.

        Returns
        -------
        labels : ndarray of shape (n_samples,)
            Index of the cluster each sample belongs to.
        """
        check_is_fitted(self)

        X = self._validate_data(X, accept_sparse='csr', reset=False,
                                dtype=[np.float64, np.float32],
                                order='C', accept_large_sparse=False)

        sample_weight = _check_sample_weight(sample_weight, X, dtype = X.dtype)
        return self.clf.predict(X)


import numpy as np
import pandas as pd
from scipy import stats
import statsmodels.api as sm
import statsmodels.formula.api as smf
import statsmodels.regression

# https://www.statsmodels.org/stable/regression.html#
# https://www.statsmodels.org/stable/generated/statsmodels.regression.linear_model.RegressionResults.html#statsmodels.regression.linear_model.RegressionResults
# https://stats.stackexchange.com/questions/16493/difference-between-confidence-intervals-and-prediction-intervals
# https://stats.stackexchange.com/questions/136157/general-mathematics-for-confidence-interval-in-multiple-linear-regression


class LinearRegressionModeler:
    def __init__(self, regressors : list = None, y_label : str = None, formula : str = None):
        if formula:
            self.formula = formula
            self.regressors = None
            self.y_label = None
        else:
            self.formula = None
            self.regressors = regressors
            self.y_label = y_label

    def fit(self, train_df : pd.DataFrame):
        if self.formula:
            self.model = smf.ols(formula = self.formula, data = train_df)
        else:
            X = train_df[self.regressors]
            Y = train_df[self.y_label]
            self.model = sm.OLS(Y, sm.add_constant(X))

        self.results = self.model.fit()

    def getResults(self) -> statsmodels.regression.linear_model.RegressionResults:
        return self.results

    def getSummary(self):
        return self.results.summary()

    def t_inv(self, alpha = 0.05):
        return stats.t.ppf(1 - alpha / 2, self.results.df_resid)

    def predict(self, test_df : pd.DataFrame):
        if self.formula:
            return self.results.predict(test_df)
        else:
            X = test_df[self.regressors]
            # has_constant: https://stackoverflow.com/questions/36532529/add-constant-in-statsmodels-not-working
            return self.results.predict(sm.add_constant(X, has_constant = 'add'))

    def predict_detail(self, test_df : pd.DataFrame, alpha : float = 0.05):
        """
        Args:
            test_df:
            alpha: 0.05 is default value, two-tail value

        Returns:
            Explanations:
            dfs = predictions.summary_frame(alpha = alpha)
            # how mean_ci is calculated:
            dfs["mean_ci_lower"] = dfs["mean"] - stats.t.ppf(1 - alpha / 2, results.df_resid) * dfs['mean_se']
            dfs["mean_ci_upper"] = dfs["mean"] + stats.t.ppf(1 - alpha / 2, results.df_resid) * dfs['mean_se']

            # how obs_ci is calculated:
            dfs["obs_ci_lower"] = dfs["mean"] - stats.t.ppf(1 - alpha / 2, results.df_resid) * np.sqrt(dfs['mean_se'] ** 2 + results.mse_resid)
        """
        if self.formula:
            predictions = self.results.get_prediction(test_df)
        else:
            X = test_df[self.regressors]
            # has_constant: https://stackoverflow.com/questions/36532529/add-constant-in-statsmodels-not-working
            predictions = self.results.get_prediction(sm.add_constant(X, has_constant = 'add'))

        summ = predictions.summary_frame(alpha = alpha)
        summ['obs_se'] = np.sqrt(summ['mean_se'] ** 2 + self.results.mse_resid)
        return summ[['mean', 'mean_se', 'obs_se', 'mean_ci_lower', 'mean_ci_upper', 'obs_ci_lower', 'obs_ci_upper']]

    @ property
    def params(self):
        return self.results.params


import contextlib
import logging

import numpy as np
import pandas as pd
from sqlalchemy import inspect, and_
from sqlalchemy.orm import Query, Session

from AqDsKit.accessors.db.dbconfigs import baseDbInf

#######################################
### Database accesor Object -- DAO  ###
#######################################

class dbIO:

    def __init__(self, dbinf: baseDbInf):
        dbinf.launch()
        self.__dbinf = dbinf
        self.placeholder = dbinf.argTempStr()

    def getDbInf(self):
        return self.__dbinf

    def getEngine(self):
        return self.getDbInf().engine

    @ contextlib.contextmanager
    def get_session(self, errormsg : str = "reason Unknown") -> Session:
        """return the session object to operate

        :param errormsg:
        :return: Session object

        example:
        >>> with self.get_session() as s:
        >>>     #do something here
        """
        session = self.getDbInf().newSession()
        try:
            yield session
            session.commit()
        except Exception as e:
            session.rollback()
            logging.error("Error Occured: %s\n%s" % (e.args, errormsg))
        finally:
            session.close()

    def insert(self, table, record : dict) -> None:
        """insert a record into db

        :param table: table class
        :param record: dict of record to insert
        :return: None

        example:
        >>> x = {"fund_id" : "160023", "date" : "2005-09-02", "net_value" : 1.0102, "full_value" : 3.0102, "div" : 0.234, "pnl" : 0.2341}
        >>> dbIO.insert(MutualFundHist, x)
        """
        with self.get_session() as s:
            s.add(table(**record))

    def update(self, table, primary_kvs : dict, record : dict) -> None:
        """update record by looking at primary key

        :param table: table class
        :param primary_kvs: dict of primary key-value pairs, use to find which record(s) to update
        :param record: dict of new record
        :return: None

        example:
        >>> p = {"fund_id" : "150234", "date" : "2005-09-09"}  # use to find which records to update_all
        >>> r = {"net_value" : 0.4456}  # the new record to save to db
        >>> dbIO.update(MutualFundHist, p, r)
        is equivalent to:
        UPDATE MutualFundHist
        SET net_value = 0.4456
        WHERE fund_id = '150234' AND date = '2005-09-09';
        """
        with self.get_session() as s:
            #s.query(table).filter(table.fund_id == "150234").update_all(r)
            conditions = [getattr(table, k) == v for k, v in primary_kvs.items()]
            s.query(table).filter(and_(*conditions)).update_all(record)

    def delete(self, table, primary_kvs : dict) -> None:
        """delete record by looking at primary key

        :param table: table class
        :param primary_kvs: dict of primary key-value pairs, use to find which record(s) to update_all
        :return: None

        example:
        >>> p = {"fund_id" : "150234", "date" : "2005-09-09"}  # use to find which records to update_all
        >>> self.delete(MutualFundHist, p)
        is equivalent to:
        DELETE FROM MutualFundHist
        WHERE fund_id = '150234' AND date = '2005-09-09';
        """
        with self.get_session() as s:
            conditions = [getattr(table, k) == v for k, v in primary_kvs.items()]
            s.query(table).filter(and_(*conditions)).delete()


    def modify_sql(self, sql : str, *args, errormsg : str = "reason Unknown") -> int:
        """execute original sql (not query)

        :param sql: sql string
        :param args: arguments to fill in sql template
        :param errormsg: error message to display when error occurred
        :return: 1: success, 0: failed

        example:
        >>> sql = 'UPDATE MutualFundHist SET net_value = %s WHERE fund_id = %s AND date = %s;'
        >>> self.modify_sql(sql, 0.7456, '150234', '2005-09-09')
        """
        with self.getEngine().connect() as conn:
            try:
                sql = sql.replace("?", self.placeholder)
                conn.execute(sql, *args)
            except Exception as e:
                logging.error("Error Occured: %s\n%s" % (e.args, errormsg))
                return 0
            else:
                return 1

    def query_df(self, query : Query) -> pd.DataFrame:
        """making query directly

        :param query: Query statement
        :return: resulting dataframe

        example:
        >>> with self.get_session() as s:
        >>>     query = s.query(MutualFundHist).filter(MutualFundHist.fund_id == "160023")
        >>>
        >>> df = self.query_df(query)
        """
        return pd.read_sql(query.statement, query.session.bind).replace({None : np.nan})

    def query_sql_df(self, sql : str, *args, errormsg : str = "reason Unknown") -> pd.DataFrame:
        """using original sql to make queries

        :param sql: the sql template
        :param args: the argument list to fill in the sql template
        :param errormsg: message to display when encoutering errors
        :return: the result dataframe

        example:
        >>> sql = "select * from mutualfundhist where fund_id = %s"
        >>> r = self.query_sql_df(sql, '160023')
        """
        with self.getEngine().connect() as conn:
            try:
                sql = sql.replace("?", self.placeholder)
                r = conn.execute(sql, *args)
                #df = pd.read_sql(sql, conn, params = args)
            except Exception as e:
                logging.error("Error Occured: %s\n%s" % (e.args, errormsg))
                df = pd.DataFrame({})
            else:
                headers = [ i[0] for i in r.cursor.description ]
                # df = pd.DataFrame.from_records(r, columns = headers)
                df = pd.DataFrame.from_records(r, columns = headers, coerce_float = True)
            finally:
                return df

    def insert_df(self, table, df : pd.DataFrame) -> None:
        """save a pandas table into database

        :param table: table class
        :param df: the pandas Dataframe to save into db
        :return:

        example:
        >>> df = pd.DataFrame({
        >>>     "fund_id" : ["160023", "160023", "160023", "150234"],
        >>>     "date" : ["2005-09-09", "2005-09-02", "2005-08-31", "2005-09-09"],
        >>>     "net_value" : [1.0234, 1.0102, 1.0456, 0.9876],
        >>>     "full_value" : [3.0234, 3.0102, 3.0456, 2.9876],
        >>>     "div" : [None, 0.234, None, None],
        >>>     "split_ratio" : [1.0034, None, None, 1.2232],
        >>>     "pnl" : [0.2334, 0.2341, -0.1442, -0.0032],
        >>> })
        >>> self.insert_df(MutualFundHist, df)
        """
        records = df2Tables(df, table)

        with self.get_session() as s:
            s.add_all(records)

    def insert_pd_df(self, tablename: str, df : pd.DataFrame, schema: str = None, mode: str = "append", index: bool = False,
                     chunksize: int = 1000, method: str = 'multi', errormsg : str = "reason Unknown") -> int:
        """save a pandas dataframe to database using pandas style

        :param tablename: the name str of the table
        :param df: the pandas dataframe to insert
        :param schema: the schema you are going to drop table into
        :param mode: {‘fail’, ‘replace’, ‘append’}, default ‘append’
        :param index: whether to include the index
        :param chunksize: Specify the number of rows in each batch to be written at a time. recommend 1000 or None
        :param method: {None, ‘multi’, callable}
        :return: 1: success, 0: failed

        check here: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_sql.html
        speed up insert:
        use method = multi will significantly speed up as it will insert in a batch
        limit chunk size in accordance with the limit of your database
        """
        try:
            with self.getEngine().begin() as conn:  # .begin() will automatically tackle with rollback() and commit()
                df.to_sql(tablename, conn, schema = schema, if_exists = mode, index = index, chunksize = chunksize, method = method)
        except Exception as e:
            logging.error("Error Occured: %s\n%s" % (e.args, errormsg))
            return 0
        else:
            return 1


    def insert_sql_df(self, tablename: str, df : pd.DataFrame, errormsg : str = "reason Unknown") -> int:
        """save a pandas dataframe to database using pythonic style

        :param tablename: the name str of the table
        :param df: the pandas dataframe to insert
        :return: 1: success, 0: failed
        """
        sql = f"""
        insert into {tablename} ({",".join(df.columns)})
        values ({",".join("?" for col in df.columns)})
        """
        sql = sql.replace("?", self.placeholder)

        conn = self.getEngine().raw_connection()
        cursor = conn.cursor()
        try:
            cursor.executemany(sql, df.values)
        except Exception as e:
            conn.rollback()
            logging.error("Error Occured: %s\n%s" % (e.args, errormsg))
            return 0
        else:
            conn.commit()
            return 1
        finally:
            cursor.close()
            conn.close()



def getPrimaryKeys(table) -> list:
    return [key.name for key in inspect(table).primary_key]

def df2Tables(df : pd.DataFrame, table) -> list:
    records = []
    for row in df.replace({np.nan: None}).to_dict(orient = "records"):
        records.append(table(**row))

    return records